unit uRtf2Html;

{$mode ObjFPC}{$H+}

interface

uses
  Classes, SysUtils, LConvEncoding, RtfPars;

type
  TRtfFontStyle = (rfsBold, rfsItalic, rfsStrikeThrough, rfsUnderline, rfsShadow, rfsOutline);
  TRtfFontStyles = set of TRtfFontStyle;

  TRtfCharPosition = (rcpNormal, rcpSubscript, rcpSuperscript);

  TRtfUnderlineStyle = (rusSingle, rusWord, rusDouble);

  TRtfCharAttrib = record
    DefaultFontNum: Integer;
    FontNum: Integer;
    FontSize: Double;
    FontStyle: TRtfFontStyles;
    Position: TRtfCharPosition;
    UnderlineStyle: TRtfUnderlineStyle;
    FgColor: Integer;
    BkColor: Integer;
  end;

  TRtfAlignment = (raLeft, raCenter, raRight, raJustify);

  TRtfParAttrib = record
    Alignment: TRtfAlignment;
    LeftIndent: Integer;  // in twips (1 inch = 1440 twips)
    RightIndent: Integer;
    FirstIndent: Integer;
    SpacingBefore: Integer;  // in twips
    SpacingAfter: Integer;
    SpacingBetween: Integer;
  end;

  TRtfCellBorder = (rcbLeft, rcbTop, rcbRight, rcbBottom);
  TRtfCellBorderStyle = (rcbsNone, rcbsSingle);
  TRtfCellBorderStyles = array[TRtfCellBorder] of TRtfCellBorderStyle;

  TRtfTblAttrib = record
    CellText: array of string;
    ColPos: array of Integer;
    CellBorders: array of TRtfCellBorderStyles;
    VertMargin: Integer;
  end;

  TRtfDocAttrib = record
  end;

  TRtf2HtmlConverter = class
  private
    FParser: TRtfParser;
    FCodePage: String;
    FTitle: String;
    FInTable: Boolean;

  protected
    // Writing output
    FOutput: TStrings;
    FCurrText: String;
    FSpanOpen: Boolean;
    FSpanDelayed: Boolean;
    FParOpen: Boolean;
    FParDelayed: Boolean;
    FFormatSettings: TFormatSettings;
    procedure Add(AString: String);
//    procedure Add(const AStrings: array of String);
    procedure AddText(const AText: String);
    procedure ClosePar;
    procedure CloseSpan;
    function Indentation: String;
    procedure WriteFooter;
    procedure WriteHeader;
    procedure WritePar(const Attr: TRtfParAttrib);
    procedure WriteSpan(const Attr: TRtfCharAttrib);
    procedure WriteTableRow(const Attr: TRtfTblAttrib);

    procedure DisplayOutput;
  protected
    // Parser handlers
    procedure DoCharAttributes;
    procedure DoCharSet;
    procedure DoCtrl;
    procedure DoDestination;
    procedure DoDocAttributes;
    procedure DoGroup;
    procedure DoParAttributes;
    procedure DoPicture;
    procedure DoSectAttributes;
    procedure DoSpecialChar;
    procedure DoTblAttributes;
    procedure DoText;
  protected
    // Attributes stack
    FStack: array of TRtfCharAttrib;
    FFontPending: Boolean;
    FIndent: Integer;
    FDocAttrib: TRtfDocAttrib;
    FCurrCharAttrib: TRtfCharAttrib;
    FParAttrib: TRtfParAttrib;
    FTblAttrib: TRtfTblAttrib;
    FCurrCellBorder: TRtfCellBorder;
    procedure ClearCharAttrib;
    procedure ClearParAttrib;
    procedure ClearTblAttrib;
    procedure PushCharAttrib;
    procedure PopCharAttrib;

  public
    constructor Create(RTFStream: TStream);
    destructor Destroy; override;
    procedure ConvertToHtml(AFileName: String);
    procedure ConvertToHtml(AStream: TStream; ATitle: String);
  end;

implementation

function rtfColorToString(C: PRtfColor; ADefaultColor: String): String;
begin
  if C <> nil then
  begin
    if (C^.rtfCBlue = -1) and (C^.rtfCGreen = -1) and (C^.rtfCRed = -1) then
      Result := ADefaultColor
    else
      Result := Format('#%.2x%.2x%.2x', [C^.rtfCRed, C^.rtfCGreen, C^.rtfCBlue]);
  end else
    Result := '';
end;

function TwipsToMM(twips: Integer): Double;
begin
  Result := twips / 1440 * 25.4;
end;

function TwipsToPts(twips: Integer): Double;
begin
  Result := twips / 20;
end;


{ TRtf2HtmlConverter }

constructor TRtf2HtmlConverter.Create(RTFStream: TStream);
begin
  inherited Create;

  FFormatSettings := DefaultFormatSettings;
  FFormatSettings.DecimalSeparator := '.';

  FIndent := 0;

  FOutput := TStringList.Create;

  FParser := TRtfParser.Create(RTFStream);
  FParser.ClassCallBacks[rtfText] := @DoText;
  FParser.ClassCallBacks[rtfGroup] := @DoGroup;
  FParser.ClassCallBacks[rtfControl] := @DoCtrl;
  FParser.DestinationCallBacks[rtfPict] := @DoPicture;
end;

destructor TRtf2HtmlConverter.Destroy;
begin
  FParser.Free;
  FOutput.Free;
  inherited;
end;

procedure TRtf2HtmlConverter.Add(AString: String);
begin
  FOutput.Add(Indentation + AString);
end;
             {
procedure TRtf2HtmlConverter.Add(const AStrings: array of String);
var
  i: Integer;
begin
  for i := Low(AStrings) to High(AStrings) do
    Add(AStrings[i]);
end;       }

procedure TRtf2HtmlConverter.AddText(const AText: String);
var
  n: Integer;
begin
  if AText = '' then
    exit;
  n := FOutput.Count;
  if n = 0 then
    FOutput.Add(AText)
  else
    FOutput[n-1] := FOutput[n-1] + AText;
end;

procedure TRtf2HtmlConverter.ClearCharAttrib;
begin
  FCurrCharAttrib.FontNum := FCurrCharAttrib.DefaultFontNum;
  FCurrCharAttrib.FontSize := 12;
  FCurrCharAttrib.FontStyle := [];
  FCurrCharAttrib.Position := rcpNormal;
  FCurrCharAttrib.UnderlineStyle := rusSingle;
  FCurrCharAttrib.FgColor := -1;
  FCurrCharAttrib.BkColor := -1;
end;

procedure TRtf2HtmlConverter.ClearParAttrib;
begin
  FParAttrib.Alignment := raLeft;
  FParAttrib.LeftIndent := 0;
  FParAttrib.RightIndent := 0;
  FParAttrib.FirstIndent := 0;
  FParAttrib.SpacingBefore := 0;
  FParAttrib.SpacingAfter := 0;
  FParAttrib.SpacingBetween := 0;
end;

procedure TRtf2HtmlConverter.ClearTblAttrib;
begin
  SetLength(FTblAttrib.CellText, 0);
  SetLength(FTblAttrib.ColPos, 0);
  SetLength(FTblAttrib.CellBorders, 0);
end;

procedure TRtf2HtmlConverter.ConvertToHtml(AFileName: String);
var
  stream: TFileStream;
begin
  stream := TFileStream.Create(AFileName, fmCreate or fmShareDenyWrite);
  try
    ConvertToHtml(stream, ExtractFileName(AFileName));
  finally
    stream.Free;
  end;
end;

procedure TRtf2HtmlConverter.ConvertToHtml(AStream: TStream; ATitle: String);
begin
  FTitle := ATitle;

  FOutput.Clear;

  FParser.StartReading;
  if FSpanOpen then
    CloseSpan;
  if FParOpen then
    ClosePar;

  FOutput.SaveToStream(AStream);
end;

procedure TRtf2HtmlConverter.DisplayOutput;
var
  i: Integer;
begin
  for i := 0 to FOutput.Count-1 do
    Writeln(FOutput[i]);
end;

procedure TRtf2HtmlConverter.DoCharAttributes;
begin
  if FCurrText <> '' then
  begin
    AddText(FCurrText);
//    FOutput.Add(FCurrText);
    FCurrText := '';
  end;

  // Push old character attributes to stack (and remember the context (major,
  // minor) in order to be able to restore the attribs in the right order).
//  PushCharAttrib;

  // Modify current character attributes.
  case FParser.rtfMinor of
    rtfPlain:
      begin
        ClearCharAttrib;
        FSpanDelayed := true;
      end;
    rtfFontNum:
      begin
        FCurrCharAttrib.FontNum := FParser.rtfParam;
        FSpanDelayed := true;
      end;
    rtfFontSize:
      begin
        FCurrCharAttrib.FontSize := FParser.rtfParam / 2.0;
        FSpanDelayed := true;
      end;
    rtfBold:       // bold font
      begin
        if FParser.rtfParam = 0 then
          FCurrCharAttrib.FontStyle := FCurrCharAttrib.FontStyle - [rfsBold]
        else
          FCurrCharAttrib.FontStyle := FCurrCharAttrib.FontStyle + [rfsBold];
        FSpanDelayed := true;
      end;
    rtfItalic:      // italic font
      begin
        if FParser.rtfParam = 0 then
          FCurrCharAttrib.FontStyle := FCurrCharAttrib.FontStyle - [rfsItalic]
        else
          FCurrCharAttrib.FontStyle := FCurrCharAttrib.FontStyle + [rfsItalic];
        FSpanDelayed := true;
      end;
    rtfUnderline,       // single underline
    rtfWUnderline:      // word underline
      begin
        if FParser.rtfParam = 0 then
          FCurrCharAttrib.FontStyle := FCurrCharAttrib.FontStyle - [rfsUnderline]
        else
          FCurrCharAttrib.FontStyle := FCurrCharAttrib.FontStyle + [rfsUnderline];
        FSpanDelayed := true;
      end;
    rtfDUnderline:  // dotted underline
      begin
        if FParser.rtfParam = 0 then
          FCurrCharAttrib.FontStyle := FCurrCharAttrib.FontStyle - [rfsUnderline]
        else
          FCurrCharAttrib.FontStyle := FCurrCharAttrib.FontStyle + [rfsUnderline];
        FSpanDelayed := true;
      end;
    rtfDbUnderline:   // double underline
      begin
        if FParser.rtfParam = 0 then
          FCurrCharAttrib.FontStyle := FCurrCharAttrib.FontStyle - [rfsUnderline]
        else
          FCurrCharAttrib.FontStyle := FCurrCharAttrib.FontStyle + [rfsUnderline];
        FSpanDelayed := true;
      end;
    rtfNoUnderline:
      begin
        FCurrCharAttrib.FontStyle := FCurrCharAttrib.FontStyle - [rfsUnderline];
        FSpanDelayed := true;
      end;
    rtfStrikeThru:     // strike-through text
      begin
        if FParser.rtfParam = 0 then
          FCurrCharAttrib.FontStyle := FCurrCharAttrib.FontStyle - [rfsStrikeThrough]
        else
          FCurrCharAttrib.FontStyle := FCurrCharAttrib.FontStyle + [rfsStrikeThrough];
        FSpanDelayed := true;
      end;
    rtfShadow:  // Shadow effect
      begin
        if FParser.rtfParam = 0 then
          FCurrCharAttrib.FontStyle := FCurrCharAttrib.FontStyle - [rfsShadow]
        else
          FCurrCharAttrib.FontStyle := FCurrCharAttrib.FontStyle + [rfsShadow];
        FSpanDelayed := true;
      end;
    rtfOutline:   // Outline effect
      begin
        if FParser.rtfParam = 0 then
          FCurrCharAttrib.FontStyle := FCurrCharAttrib.FontSTyle - [rfsOutline]
        else
          FCurrCharAttrib.FontStyle := FCurrCharAttrib.FontSTyle + [rfsOutline];
        FSpanDelayed := true;
      end;
    rtfSubScript:
      begin
        if FParser.rtfParam = 0 then
          FCurrCharAttrib.Position := rcpNormal
        else
          FCurrCharAttrib.Position := rcpSubscript;
        FSpanDelayed := true;
      end;
    rtfSuperScript:
      begin
        if FParser.rtfParam = 0 then
          FCurrCharAttrib.Position := rcpNormal
        else
          FCurrCharAttrib.Position := rcpSuperscript;
        FSpanDelayed := true;
      end;
    rtfNoSuperSub:
      begin
        FCurrCharAttrib.Position := rcpNormal;
        FSpanDelayed := true;
      end;
    rtfForeColor:
      begin
        FCurrCharAttrib.FgColor := FParser.rtfParam;
        FSpanDelayed := true;
      end;
    rtfBackColor:
      begin
        FCurrCharAttrib.BkColor := FParser.rtfParam;
        FSpanDelayed := true;
      end;
  end;
end;

procedure TRtf2HtmlConverter.DoCharSet;
begin
  case FParser.rtfMinor of
    rtfMacCharSet:
      FCodePage := encodingCPMac;
    rtfAnsiCharSet:
      FCodePage := encodingCP1252;
    rtfPCCharSet:
      FCodePage := encodingCP437;
    rtfPcaCharSet:
      FCodePage := encodingCP850;
    otherwise
      FCodePage := encodingCP1252;
      // there is also a \ansicpgN, but it is not supported by RTFPars
  end;
end;

procedure TRtf2HtmlConverter.DoCtrl;
begin
  case FParser.rtfMajor of
    rtfCharAttr:
      DoCharAttributes;
    rtfCharSet:
      DoCharset;
    rtfDefFont:
      FCurrCharAttrib.DefaultFontNum := FParser.rtfParam;
    rtfDestination:
      DoDestination;
    rtfDocAttr:
      DoDocAttributes;
    rtfSpecialChar:
      DoSpecialChar;
    rtfParAttr:
      DoParAttributes;
    rtfSectAttr:
      DoSectAttributes;
    rtfTblAttr:
      DoTblAttributes;
  end;
end;

procedure TRtf2HtmlConverter.DoDestination;
begin
  case FParser.rtfMajor of
    rtfPict:
      writeln('pict');
  end;
end;

procedure TRtf2HtmlConverter.DoDocAttributes;
begin
  {
  case FParser.rtfMinor of
  end;
  }
end;

procedure TRtf2HtmlConverter.DoGroup;
begin
  case FParser.rtfMajor of
    rtfBeginGroup:
      begin
        inc(FIndent, 2);
        AddText(FCurrText);
        FCurrText := '';
        PushCharAttrib;
      end;
    rtfEndGroup:
      begin
        if FCurrText <> '' then
        begin
          AddText(FCurrText);
          FCurrText := '';
          if FSpanOpen then
            CloseSpan;
        end;
        PopCharAttrib;
        dec(FIndent, 2);
      end;
  end;
end;

procedure TRtf2HtmlConverter.DoParAttributes;
var
  n: Integer;
begin
  case FParser.RtfMinor of
    rtfParDef:    // \pard -- begins a paragraph and defines its properties
      begin
        if FSpanOpen then
          CloseSpan;
        if FParOpen then
          ClosePar;
        ClearParAttrib;
        FParDelayed := true;
      end;
    rtfQuadLeft:   // \ql -- left-aligned paragraph
      begin
        FParAttrib.Alignment := raLeft;
        FParDelayed := true;
      end;
    rtfQuadRight:  // \qr -- right-aligned paragraph
      begin
        FParAttrib.Alignment := raRight;
        FParDelayed := true;
      end;
    rtfQuadCenter:   // \qc -- centered paragraph
      begin
        FParAttrib.Alignment := raCenter;
        FParDelayed := true;
      end;
    rtfQuadJust:    // \qj -- justified paragrah
      begin
        FParAttrib.Alignment := raJustify;
        FParDelayed := true;
      end;
    rtfLeftIndent:
      begin
        FParAttrib.LeftIndent := FParser.rtfParam;
        FParDelayed := true;
      end;
    rtfRightIndent:
      begin
        FParAttrib.RightIndent := FParser.rtfParam;
        FParDelayed := true;
      end;
    rtfFirstIndent:
      begin
        FParAttrib.FirstIndent := FParser.rtfParam;
        FParDelayed := true;
      end;
    rtfSpaceAfter:
      begin
        FParAttrib.SpacingAfter := FParser.rtfParam;
        FParDelayed := true;
      end;
    rtfSpaceBefore:
      begin
        FParAttrib.SpacingBefore := FParser.rtfParam;
        FParDelayed := true;
      end;
    rtfSpaceBetween:
      begin
        FParAttrib.SpacingBetween := FParser.rtfParam;
        FParDelayed := true;
      end;

    // Table-related
    rtfInTable:      // \intbl - signals that paragraph belongs to a table
      begin
        FInTable := true;
      end;
    rtfBorderSingle:  // Single cell border
      begin
        n := Length(FTblAttrib.CellBorders);
        SetLength(FTblAttrib.CellBorders, n+1);
        FTblAttrib.CellBorders[n, FCurrCellBorder] := rcbsSingle;
      end;
  end;
end;

procedure TRtf2HtmlConverter.DoPicture;
begin
 // WriteLn('PCITRE');
end;

procedure TRtf2HtmlConverter.DoSectAttributes;
begin
  {
  case FParser.RtfMinor of
  end;
  }
end;

procedure TRtf2HtmlConverter.DoSpecialChar;
var
  c: WideChar;
  n: Integer;
begin
  case FParser.RtfMinor of
    rtfLine:            // \line -- new line
      FCurrText := FCurrText + '<br />'; //AddText('<br />');
    rtfNoBrkSpace:      // \~ -- non-breaking space
      FCurrText := FCurrText + '&nbsp;';
//      AddText('&nbsp;');
    rtfPar:             // \par -- new paragraph with settings as previous one
      begin
        // Write any collected text
        AddText(FCurrText);
        FCurrText := '';
        if FSpanOpen then
          CloseSpan;
        if FParOpen then
          ClosePar;
        FParDelayed := true;
        FSpanDelayed := true;
      end;
    rtfOptDest:         // \* --> ignore till end of group
      FParser.SkipGroup;
    rtfTab:
      FCurrText := FCurrText + '&#9;';   // but: is not rendered in html
    rtfUnicodeID:
      begin
        c := WideChar(FParser.rtfParam);
        FCurrText := FCurrText + UTF8Encode(c);
      end;

    // Table-related
    rtfCell:  // Ends a table cell
      begin
        n := Length(FTblAttrib.CellText);
        SetLength(FTblAttrib.CellText, n+1);
        FTblAttrib.CellText[n] := FCurrText;
        FCurrText := '';
      end;
    rtfRow:  // Ends a table row
      WriteTableRow(FTblAttrib);
  end;
end;

procedure TRtf2HtmlConverter.DoTblAttributes;
var
  n: Integer;
begin
  case FParser.RtfMinor of
    rtfRowDef:   // Begins a new row
      begin
        if FCurrText <> '' then
        begin
          if FSpanDelayed then
          begin
            WriteSpan(FCurrCharAttrib);
            FSpanDelayed := false;
            FSpanOpen := true;
          end;
          if FSpanOpen then
            CloseSpan;
          FCurrText := '';
        end;
        ClearTblAttrib;
      end;
    rtfCellPos:          // \cellx
      begin
        n := Length(FTblAttrib.ColPos);
        SetLength(FTblAttrib.ColPos, n+1);
        FTblAttrib.ColPos[n] := FParser.rtfParam;
      end;
    rtfCellBordLeft:     // \clbrdrl
      FCurrCellBorder := rcbLeft;
    rtfCellBordTop:      // \clbrdrt
      FCurrCellBorder := rcbTop;
    rtfCellBordRight:    // \clbrdrr
      FCurrCellBorder := rcbRight;
    rtfCellBordBottom:   // clbrdrb
      FCurrCellBorder := rcbBottom;
    rtfRowGapH:          // trgaph
      FTblAttrib.VertMargin := FParser.rtfParam;
  end;
end;

procedure TRtf2HtmlConverter.DoText;
var
  c: Char;
  s: String;
begin
  if (FCurrText = '') then
  begin
    // This is the very first character -- we must write the HTML header.
    if FOutput.Count = 0 then
      WriteHeader;

    if FSpanOpen then
      CloseSpan;
  end;

  c := chr(FParser.RTFMajor);
  if c = #0 then   // last character
  begin
    WriteFooter;
    exit;
  end;

  if c > #127 then
  begin
    s := ConvertEncoding(c, FCodePage, encodingUTF8);
    FCurrText := FCurrText + s;
  end else
    FCurrText := FCurrText + c;

  if FParDelayed then
  begin
    WritePar(FParAttrib);
    FParDelayed := false;
    FParOpen := true;
  end;

  if FSpanDelayed then
  begin
    WriteSpan(FCurrCharAttrib);
    FSpanDelayed := false;
    FSpanOpen := true;
  end;
end;

function TRtf2HtmlConverter.Indentation: String;
var
  i: Integer;
begin
  SetLength(Result, FIndent);
  if FIndent > 0 then
    FillChar(Result[1], FIndent, ' ');
end;

procedure TRtf2HtmlConverter.PopCharAttrib;
var
  n: Integer;
begin
  n := Length(FStack) - 1;
  if n <= 0 then
    exit;
  FCurrCharAttrib := FStack[n];
  SetLength(FStack, n-1);
end;

procedure TRtf2HtmlConverter.PushCharAttrib;
var
  n: Integer;
begin
  n := Length(FStack);
  SetLength(FStack, n+1);
  FStack[n] := FCurrCharAttrib;
end;

procedure TRtf2HtmlConverter.ClosePar;
begin
  dec(FIndent, 2);
  Add('</p>');
  FParOpen := false;
end;

procedure TRtf2HtmlConverter.CloseSpan;
begin
  AddText('</span>');
  FSpanOpen := false;
end;

procedure TRtf2HtmlConverter.WriteFooter;
begin
  FIndent := 2;
  Add(  '</body>');
  FIndent := 0;
  Add('</html>');
end;

procedure TRtf2HtmlConverter.WriteHeader;
var
  defaultfont: String = '';
  fontSizeValue: Double;
  font: PRTFFont;
begin
  Add('<html>');
  inc(FIndent, 2);
  Add(  '<head>');
  inc(FIndent, 2);
  Add(    '<meta charset="utf-8">');
  Add(    '<title>' + FTitle + '</title>');
  dec(Findent, 2);
  Add(  '</head>');

  font := FParser.Fonts[FCurrCharAttrib.DefaultFontNum];
  if Assigned(font) then
  begin
    fontSizeValue := FCurrCharAttrib.FontSize;
    if fontSizeValue = 0 then
      fontSizeValue := 12;
    defaultFont := Format('font-family:%s; font-size:%.1fpt;', [
       font^.rtfFName, fontSizeValue
       ], FFormatSettings
    );
    Add(Format('<body style="%s">', [defaultFont]));
  end else
    Add('<body>');
  inc(FIndent, 2);
end;

procedure TRtf2HtmlConverter.WritePar(const Attr: TRtfParAttrib);
var
  align: String = '';
  leftindent: String = '';
  rightindent: String = '';
  firstindent: String = '';
  spaceBefore: String = '';
  spaceAfter: String = '';
  spaceBetween: String = '';
  pre: String = '';
begin
  case Attr.Alignment of
    raLeft: align := 'text-align:left;';
    raCenter: align := 'text-align:center;';
    raRight: align := 'text-align:right;';
    raJustify: align := 'text-align:justify;';
  end;
  leftindent := Format('margin-left:%.1fpt;', [TwipsToPts(Attr.LeftIndent)], FFormatSettings);
  rightindent := Format('margin-right:%.1fpt;', [TwipsToPts(Attr.RightIndent)], FFormatSettings);
  firstindent := Format('text-indent:%.1fpt;', [TwipsToPts(Attr.FirstIndent)], FFormatSettings);
  spaceBefore := Format('margin-top:%.1fpt;', [TwipsToPts(Attr.SpacingBefore)], FFormatSettings);
  spaceAfter := Format('margin-bottom:%.1fpt;', [TwipsToPts(Attr.SpacingAfter)], FFormatSettings);
  if Attr.SpacingBetween > 0 then
    spaceBetween := Format('line-height:%.1fpt;', [TwipsToPts(Attr.SpacingBetween)], FFormatSettings);
  //pre := 'white-space:pre;';       // should be used by code paragraphs. Don't know how to identify them...
  Add(Format('<p style="%s%s%s%s%s%s%s%s">', [
    align, leftindent, rightindent, firstindent,
    spaceBefore, spaceAfter, spaceBetween, pre
  ]));
  inc(FIndent, 2);
  Add('');
end;

procedure TRtf2HtmlConverter.WriteSpan(const Attr: TRtfCharAttrib);
var
  fontFamily: String;
  fontSize: String;
  fontColor: String = '';
  fontStyle: String = '';
  bkColor: String = '';
  position: String = '';
  rtfColor: PRtfColor;
  fontSizeValue: Double;
begin
  fontFamily := Format('font-family:%s;', [FParser.Fonts[Attr.FontNum]^.rtfFName]);

  fontSizeValue := Attr.FontSize;
  if fontSizeValue = 0 then
    fontSizeValue := 12;
  if Attr.Position <> rcpNormal then
    fontSizeValue := fontSizeValue * 0.67;
  fontSize := Format('font-size:%.1fpt;', [fontSizeValue], FFormatSettings);

  if Attr.FgColor <> -1 then
    fontColor := Format('color:%s;', [rtfColorToString(FParser.Colors[Attr.FgColor], 'black')]);

  if Attr.BkColor <> -1 then
  begin
    rtfColor := FParser.Colors[Attr.BkColor];
    bkColor := Format('background-color:%s;', [rtfColorToString(rtfColor, 'white')]);
  end;

  fontStyle := '';
  if (rfsBold in Attr.FontStyle) then
    fontStyle := fontStyle + 'font-weight:bold;';
  if (rfsItalic in Attr.FontStyle) then
    fontStyle := fontStyle + 'font-style:italic;';
  if (rfsUnderline in Attr.FontStyle) then
    case Attr.UnderlineStyle of
      rusSingle,
      rusWord: fontStyle := fontStyle + 'text-decoration:underline;';
      rusDouble: fontStyle := fontStyle + 'text-decoration-line:underline;text-decoration-style:double;';
    end;
  if (rfsStrikeThrough in Attr.FontStyle) then
    fontStyle := fontStyle + 'text-decoration:line-through;';
  if (rfsShadow in Attr.FontStyle) then
    fontStyle := fontStyle + 'text-shadow:2px 2px #CCCCCC;';
  if (rfsOutline in Attr.FontStyle) then
    fontStyle := fontStyle + 'text-shadow: -1px 1px 2px #000, 1px 1px 2px #000, 1px -1px 0px #000, -1px -1px 0px #000;color:white;';
       //'text-shadow: -1px 1px 2px #000, 1px 2px 6px #000, 1px -1px 0px #000, -1px -1px 0px #000;',  // https://www.educative.io/answers/how-to-create-a-text-outline-using-css

  case Attr.Position of
    rcpNormal: position := 'vertical-align:baseline;';
    rcpSubscript: position := 'vertical-align:sub;';
    rcpSuperscript: position := 'vertical-align:super;';
  end;

  AddText(Format('<span style="%s%s%s%s%s%s">', [fontFamily, fontSize, fontColor, fontStyle, bkColor, position]));
end;

procedure TRtf2HtmlConverter.WriteTableRow(const Attr: TRtfTblAttrib);
var
  rowStr: String;
  widthStr: String;
  borderStr: String = '';
  marginStr: String = '';
  c: Integer;
  colWid: array of Integer = nil;
  border: TRtfCellBorderStyle;
  b: TRtfCellBorder;
  n: Integer;
begin
  // Prepare the colum widths
  n := Length(Attr.ColPos);
  SetLength(colWid, n);
  if n > 0 then
  begin
    colWid[0] := Attr.ColPos[0];
    for c := 1 to n-1 do
      colWid[c] := Attr.ColPos[c] - colWid[c-1];
  end;
  if n < Length(Attr.CellText) then
  begin
    SetLength(colWid, Length(Attr.CellText));
    for c := n to High(colWid) do
      colWid[c] := 2*1440;
  end;

  // Collapse all borders (all cells, left/top/right/bottom borders) into a single style
  border := rcbsNone;
  for c := 0 to High(Attr.CellBorders) do
  begin
    for b in TRtfCellBorder do
      if Attr.CellBorders[c, b] <> rcbsNone then
      begin
        border := Attr.CellBorders[c, b];
        break;
      end;
    if border <> rcbsNone then break;
  end;

  //marginStr := Format('padding-top:%0:.1fmm;padding-bottom:%0:.1fmm;', [TwipsToMM(Attr.VertMargin)], FFormatSettings);

  // Create html text
  rowStr := Format('<table><tr style="%s">', [marginStr]);
  for c := 0 to High(Attr.CellText) do
  begin
    widthStr := Format('width:%.1fmm;', [TwipsToMM(colWid[c])], FFormatSettings);
    case border of
      rcbsSingle:
        borderStr := 'border:1px solid black;border-collapse:collapse;';
      else
        borderStr := '';
    end;
    rowStr := Format('%s<td style="%s%s">%s</td>', [
      rowStr, widthStr, borderStr, Attr.CellText[c]
    ]);
  end;
  rowstr := rowStr + '</tr></table>';
  Add(rowstr);
end;

end.

